# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Blob Reader Node

This is a SQL Query Execution Plan Node.

This Node creates datasets based on function calls like VALUES and UNNEST.
"""
import random
import time

from typing import Iterable

import pyarrow

from opteryx.models import Columns, QueryProperties
from opteryx.managers.expression import NodeType, evaluate
from opteryx.operators import BasePlanNode
from opteryx.exceptions import SqlError
from opteryx.utils import arrays


def _generate_series(alias, *args):
    value_array = arrays.generate_series(*args)
    return [{alias: value} for value in value_array]


def _unnest(alias, values):
    """unnest converts an list into rows"""
    list_items = values.value
    if values.token_type == NodeType.NESTED:
        # single item lists are reported as nested
        from opteryx.samples import no_table

        list_items = evaluate(values, no_table(), True)
    return [{alias: row} for row in list_items]


def _values(alias, *values):
    return values


def _fake_data(alias, *args):
    rows, columns = int(args[0].value), int(args[1].value)
    return [
        {f"column_{col}": random.getrandbits(16) for col in range(columns)}
        for row in range(rows)
    ]


FUNCTIONS = {
    "fake": _fake_data,
    "generate_series": _generate_series,
    "unnest": _unnest,
    "values": _values,
}


class FunctionDatasetNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        """
        The Blob Reader Node is responsible for reading the relevant blobs
        and returning a Table/Relation.
        """
        super().__init__(properties=properties)
        self._alias = config["alias"]
        self._function = config["dataset"]["function"]
        self._args = config["dataset"]["args"]

    @property
    def config(self):  # pragma: no cover
        if self._alias:
            return f"{self._function} => {self._alias}"
        return f"{self._function}"

    @property
    def name(self):  # pragma: no cover
        return "Dataset Constructor"

    def execute(self) -> Iterable:

        try:
            start_time = time.time_ns()
            data = FUNCTIONS[self._function](self._alias, *self._args)  # type:ignore
            self.statistics.time_data_read += time.time_ns() - start_time
        except TypeError as err:  # pragma: no cover
            if str(err).startswith("_unnest() takes 2"):
                raise SqlError(
                    "UNNEST expects a literal list in paranthesis, or a field name as a parameter."
                )
            raise err

        table = pyarrow.Table.from_pylist(data)

        self.statistics.rows_read += table.num_rows
        self.statistics.columns_read += len(table.column_names)

        table = Columns.create_table_metadata(
            table=table,
            expected_rows=table.num_rows,
            name=self._function,
            table_aliases=[self._alias],
        )
        yield table



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Collection Reader Node

This is a SQL Query Execution Plan Node.

This Node primarily is used for reading NoSQL sources like MongoDB and Firestore.
"""
import time

from typing import Iterable, Optional

import pyarrow

from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.models.columns import Columns


class CollectionReaderNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        """
        The Collection Reader Node is responsible for reading the relevant documents
        from a NoSQL document store and returning a Table/Relation.
        """
        super().__init__(properties=properties)

        self._alias = config.get("alias")

        dataset = config["dataset"]
        self._dataset = ".".join(dataset.split(".")[:-1])
        self._collection = dataset.split(".")[0]

        self._reader = config.get("reader")()

        # pushed down selection/filter
        self._selection = config.get("selection")

    @property
    def config(self):  # pragma: no cover
        if self._alias:
            return f"{self._dataset} => {self._alias}"
        return f"{self._dataset}"

    @property
    def name(self):  # pragma: no cover
        return "Collection Reader"

    def execute(self) -> Iterable:

        metadata = None

        row_count = self._reader.get_document_count(self._collection)

        for page in self._reader.read_documents(self._collection):

            start_read = time.time_ns()
            pyarrow_page = pyarrow.Table.from_pylist(page)

            self.statistics.time_data_read += time.time_ns() - start_read
            self.statistics.rows_read += pyarrow_page.num_rows
            self.statistics.bytes_processed_data += pyarrow_page.nbytes
            self.statistics.document_pages += 1

            if metadata is None:
                pyarrow_page = Columns.create_table_metadata(
                    table=pyarrow_page,
                    expected_rows=row_count,
                    name=self._dataset,
                    table_aliases=[self._alias],
                )
                metadata = Columns(pyarrow_page)
                self.statistics.collections_read += 1
            else:
                pyarrow_page = metadata.apply(pyarrow_page)

            yield pyarrow_page



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This is a persisted JSON document as a KV store - this is a really bad idea, but
exists as a KV store of last resort. This is probably only going to be used on
Raspberry Pis and toy implementations of Opteryx.
"""

import base64
import io

import orjson

from opteryx.managers.kvstores import BaseKeyValueStore


class LocalKVJson(BaseKeyValueStore):
    def __init__(self, location):
        self._location = location

    def get(self, key):
        if hasattr(key, "decode"):
            key = key.decode()
        try:
            with open(self._location, mode="rb") as f:
                doc = orjson.loads(f.read())
                value = doc.get(key)
                if value:
                    value = base64.b85decode(value)
                    return io.BytesIO(value)
                return None
        except FileNotFoundError:
            return None

    def put(self, key, value):
        # for comatibility with RocksDB
        self.set(key, value)

    def set(self, key, value):
        if hasattr(key, "decode"):
            key = key.decode()

        if not value:
            bytes_value = None
        else:
            value.seek(0, 0)
            bytes_value = base64.b85encode(value.read()).decode()
            value.seek(0, 0)
        try:
            with open(self._location, mode="rb") as f:
                doc = orjson.loads(f.read())
        except FileNotFoundError:
            doc = {}
        doc.pop(key, None)
        if bytes_value:
            doc[key] = bytes_value
        with open(self._location, mode="wb") as f:
            f.write(orjson.dumps(doc))

    def contains(self, keys):
        if not isinstance(keys, (list, set, tuple)):
            keys = [keys]
        keys = [key.decode() if hasattr(key, "decode") else key for key in keys]
        try:
            with open(self._location, mode="r") as f:
                doc = orjson.loads(f.read())
            return [k for k in doc.keys() if k in keys]
        except FileNotFoundError:
            return []



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Selection Node

This is a SQL Query Execution Plan Node.
"""
import time

from typing import Iterable

import numpy
import pyarrow

from pyarrow import Table

from opteryx.attribute_types import TOKEN_TYPES
from opteryx.exceptions import SqlError
from opteryx.managers.expression import evaluate
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode


class SelectionNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.filter = config.get("filter")
        self._unfurled_filter = None
        self._mapped_filter = None

    @property
    def config(self):  # pragma: no cover
        def _inner_config(predicate):
            if isinstance(predicate, tuple):
                if len(predicate) > 1 and predicate[1] == TOKEN_TYPES.IDENTIFIER:
                    return f"`{predicate[0]}`"
                if len(predicate) > 1 and predicate[1] == TOKEN_TYPES.VARCHAR:
                    return f'"{predicate[0]}"'
                if len(predicate) == 2:
                    if predicate[0] == "Not":
                        return f"NOT {_inner_config(predicate[1])}"
                    return f"{predicate[0]}"
                return "(" + " ".join(_inner_config(p) for p in predicate) + ")"
            if isinstance(predicate, list):
                if len(predicate) == 1:
                    return _inner_config(predicate[0])
                return "[" + ",".join(_inner_config(p) for p in predicate) + "]"
            return f"{predicate}"

        return _inner_config(self.filter)

    @property
    def name(self):  # pragma: no cover
        return "Selection"

    def execute(self) -> Iterable:

        if len(self._producers) != 1:
            raise SqlError(f"{self.name} on expects a single producer")

        data_pages = self._producers[0]  # type:ignore
        if isinstance(data_pages, Table):
            data_pages = (data_pages,)

        schema = None
        at_least_one = False

        # we should always have a filter - but no harm in checking
        if self.filter is None:
            yield from data_pages

        else:

            for page in data_pages.execute():

                if schema is None:
                    schema = page.schema

                start_selection = time.time_ns()
                mask = evaluate(self.filter, page, False)
                self.statistics.time_evaluating += time.time_ns() - start_selection

                # if the mask is a boolean array, we've called a function that
                # returns booleans
                if isinstance(mask, pyarrow.lib.BooleanArray) or (
                    isinstance(mask, numpy.ndarray) and mask.dtype == numpy.bool_
                ):
                    mask = numpy.nonzero(mask)[0]

                self.statistics.time_selecting += time.time_ns() - start_selection

                # if there's no matching rows, just drop the page
                if mask.size > 0:
                    yield page.take(mask)
                    at_least_one = True

        # we need to send something to the next operator, send an empty table
        if not at_least_one:
            yield pyarrow.Table.from_arrays([[] for i in schema.names], schema=schema)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy

from pyarrow import compute


def string_slicer_left(arr, length):
    """
    Slice a list of strings from the left
    """
    if len(arr) == 0:
        return [[]]
    if hasattr(length, "__iter__"):  # [#325]
        length = length[0]
    length = int(length)
    if length == 0:
        return [[""] * len(arr)]
    return [None if s is None else s[:length] for s in arr]


def string_slicer_right(arr, length):
    """
    Slice a list of strings from the right
    """
    if len(arr) == 0:
        return [[]]
    if hasattr(length, "__iter__"):  # [#325]
        length = length[0]
    length = int(length)
    if length == 0:
        return [[""] * len(arr)]
    return [None if s is None else s[-length:] for s in arr]


def soundex(arr):
    from opteryx.third_party.fuzzy.soundex import Soundex

    _soundex = Soundex(4)
    interim = ["0000"] * arr.size

    for index, string in enumerate(arr):
        if string:
            interim[index] = _soundex(string)
        else:
            interim[index] = None

    return numpy.array(interim, dtype=numpy.str_)


def get_md5(item):
    """calculate MD5 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.md5(str(item).encode()).hexdigest()  # nosec - meant to be MD5


def get_sha1(item):
    """calculate SHA1 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha1(str(item).encode()).hexdigest()


def get_sha224(item):
    """calculate SHA256 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha224(str(item).encode()).hexdigest()


def get_sha256(item):
    """calculate SHA256 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha256(str(item).encode()).hexdigest()


def get_sha384(item):
    """calculate SHA256 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha384(str(item).encode()).hexdigest()


def get_sha512(item):
    """calculate SHA512 hash of a value"""
    import hashlib  # delay the import - it's rarely needed

    if item is None:
        return None

    return hashlib.sha512(str(item).encode()).hexdigest()


def get_base64_encode(item):
    """calculate BASE64 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b64encode(item).decode("UTF8")


def get_base64_decode(item):
    """calculate BASE64 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b64decode(item).decode("UTF8")


def get_base85_encode(item):
    """calculate BASE85 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b85encode(item).decode("UTF8")


def get_base85_decode(item):
    """calculate BASE85 encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b85decode(item).decode("UTF8")


def get_hex_encode(item):
    """calculate HEX encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b16encode(item).decode("UTF8")


def get_hex_decode(item):
    """calculate HEX encoding of a string"""
    import base64

    if item is None:
        return None

    if not isinstance(item, bytes):
        item = str(item).encode()
    return base64.b16decode(item).decode("UTF8")


def concat(list_values):
    """concatenate a list of strings"""
    result = []
    for row in list_values:
        if row is None:
            result.append(None)
        else:
            row = row.astype(dtype=numpy.str_)
            result.append("".join(row))
    return result


def concat_ws(separator, list_values):
    """concatenate a list of strings with a separator"""
    result = []
    if len(separator) > 0:
        separator = separator[0]
        if separator is None:
            return None
    for row in list_values:
        if row is None:
            result.append(None)
        else:
            row = row.astype(dtype=numpy.str_)
            result.append(separator.join(row))
    return result


def starts_w(arr, test):
    return compute.starts_with(arr, test[0])


def ends_w(arr, test):
    return compute.ends_with(arr, test[0])


def substring(arr, from_pos, count):

    if len(arr) == 0:
        return [[]]

    def _inner(val, _from, _for):
        _from = int(_from)
        if _from > 0:
            _from -= 1
        _for = int(_for) if _for == _for else None  # nosec
        if _for is None:
            return val[_from:]
        return val[_from : _for + _from]

    def _outer():
        for index, value in enumerate(arr):
            yield _inner(value, from_pos[index], count[index])

    return list(_outer())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Global Buffer Pool.

This is little more than a wrapper around the LRU-K(2) cache.
"""
import io

from opteryx.utils.lru_2 import LRU2


class _BufferPool:
    """
    This is the class that there is a single instance of.

    This is the Buffer Pool, currently implemented as a dictionary of buffers.
    """

    slots = "_lru"

    def __init__(self, size):
        self._lru = LRU2(size=size)

    def get(self, key, cache):
        """Retrieve an item from the pool, return None if the item isn't found"""
        value = self._lru.get(key)
        if value is not None:
            return io.BytesIO(value)
        if cache is not None:
            value = cache.get(key)
        return value

    def set(self, key, value, cache):
        """Put an item into the pool, evict an item if the pool is full"""
        value.seek(0, 0)
        evicted = self._lru.set(key, value.read())
        value.seek(0, 0)
        if cache is not None:
            cache.set(key, value)
        return evicted

    @property
    def stats(self):
        """hit, miss and eviction statistics"""
        return self._lru.stats

    def reset(self, reset_stats: bool = False):
        """reset the statistics"""
        self._lru.reset(reset_stats=reset_stats)


class BufferPool(_BufferPool):

    _kv = None

    def __new__(cls):
        if cls._kv is None:
            # import here to avoid cicular imports
            from opteryx import config

            local_buffer_pool_size = config.LOCAL_BUFFER_POOL_SIZE
            cls._kv = _BufferPool(size=local_buffer_pool_size)
        return cls._kv



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Offset Node

This is a SQL Query Execution Plan Node.

This Node skips over tuples.
"""
from typing import Iterable

import pyarrow

from opteryx.exceptions import SqlError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.models import Columns


class ShowCreateNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._table = config.get("table")

    @property
    def name(self):  # pragma: no cover
        return "ShowCreate"

    @property
    def config(self):  # pragma: no cover
        return str(self._offset)

    def execute(self) -> Iterable:

        statement = f"CREATE TABLE `{self._table}` (\n"

        if len(self._producers) != 1:
            raise SqlError(f"{self.name} on expects a single producer")

        data_pages = self._producers[0]  # type:ignore
        if isinstance(data_pages, pyarrow.Table):
            data_pages = (data_pages,)

        page = next(data_pages.execute())

        columns = Columns(page)
        preferred_names = columns.preferred_column_names
        column_names = []
        for col in page.column_names:
            column_names.append([c for a, c in preferred_names if a == col][0])
        page = page.rename_columns(column_names)

        for column in page.column_names:
            column_data = page.column(column)
            column_type = str(column_data.type)
            statement += f"\t{column.ljust(32)} {column_type},\n"
        statement += ");"

        table = pyarrow.Table.from_pylist([{"show_table_create": statement}])
        table = Columns.create_table_metadata(
            table=table,
            expected_rows=1,
            name="show_table_create",
            table_aliases=[],
        )
        yield table



import numpy

from opteryx.exceptions import SqlError
from opteryx.utils import intervals, dates


def generate_series(*args):

    from opteryx.managers.expression import NodeType

    arg_len = len(args)
    arg_vals = [i.value for i in args]
    first_arg_type = args[0].token_type

    # if the parameters are numbers, generate series is an alias for range
    if first_arg_type in (NodeType.LITERAL_NUMERIC, numpy.float64):
        if arg_len not in (1, 2, 3):
            raise SqlError("generate_series for numbers takes 1,2 or 3 parameters.")
        return intervals.generate_range(*arg_vals)

    # if the params are timestamps, we create time intervals
    if first_arg_type == NodeType.LITERAL_TIMESTAMP:
        if arg_len != 3:
            raise SqlError(
                "generate_series for dates needs start, end, and interval parameters"
            )
        return dates.date_range(*arg_vals)

    # if the param is a CIDR, we create network ranges
    if first_arg_type == NodeType.LITERAL_VARCHAR:
        if arg_len not in (1,):
            raise SqlError("generate_series for strings takes 1 CIDR parameter.")

        import ipaddress

        ips = ipaddress.ip_network(arg_vals[0], strict=False)
        return [str(ip) for ip in ips]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Inner Join Node

This is a SQL Query Execution Plan Node.

This performs a INNER JOIN
"""
from typing import Iterable

import numpy
import pyarrow

from opteryx.managers.expression import NodeType
from opteryx.operators import BasePlanNode
from opteryx.models import Columns, QueryProperties
from opteryx.exceptions import ColumnNotFoundError, SqlError
from opteryx.third_party import pyarrow_ops
from opteryx.utils import arrow

INTERNAL_BATCH_SIZE = 500  # config


def calculate_batch_size(cardinality):
    """dynamically work out the processing batch size for the USING JOIN"""
    # - HIGH_CARDINALITY_BATCH_SIZE (over 90% unique) = INTERNAL_BATCH_SIZE
    # - MEDIUM_CARDINALITY_BATCH_SIZE (5% > n < 90%) = INTERNAL_BATCH_SIZE * n
    # - LOW_CARDINALITY_BATCH_SIZE (less than 5% unique) = 5
    # These numbers have had very little science put into them, they are unlikely
    # to be optimal
    if cardinality < 0.05:
        return 5
    if cardinality > 0.9:
        return INTERNAL_BATCH_SIZE
    return INTERNAL_BATCH_SIZE * cardinality


def get_columns(expression, left_columns, right_columns):

    left = []
    right = []

    if expression.token_type == NodeType.AND:
        left_left, left_right = get_columns(
            expression.left, left_columns, right_columns
        )
        right_left, right_right = get_columns(
            expression.right, left_columns, right_columns
        )
        left.extend(left_left)
        left.extend(right_left)
        right.extend(left_right)
        right.extend(right_right)
    elif (
        expression.token_type == NodeType.COMPARISON_OPERATOR
        and expression.value == "Eq"
    ):
        try:
            right = [
                right_columns.get_column_from_alias(
                    expression.right.value, only_one=True
                )
            ]
            left = [
                left_columns.get_column_from_alias(expression.left.value, only_one=True)
            ]
        except ColumnNotFoundError:
            # the ON condition may not always be in the order of the tables
            # purposefully reference the values the wrong way around
            right = [
                right_columns.get_column_from_alias(
                    expression.left.value, only_one=True
                )
            ]
            left = [
                left_columns.get_column_from_alias(
                    expression.right.value, only_one=True
                )
            ]
    else:
        raise SqlError("JOIN 'on' condition can only be comprised of 'AND's and '='s.")

    return left, right


class InnerJoinNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._right_table = config.get("right_table")
        self._join_type = config.get("join_type", "CrossJoin")
        self._on = config.get("join_on")
        self._using = config.get("join_using")

    @property
    def name(self):  # pragma: no cover
        return "Inner Join"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:

        if len(self._producers) != 2:
            raise SqlError(f"{self.name} expects two producers")

        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        self._right_table = pyarrow.concat_tables(
            right_node.execute(), promote=True
        )  # type:ignore

        if self._using:

            right_columns = Columns(self._right_table)
            left_columns = None
            right_join_columns = [
                right_columns.get_column_from_alias(col, only_one=True)
                for col in self._using
            ]

            for page in left_node.execute():

                if left_columns is None:
                    left_columns = Columns(page)
                    left_join_columns = [
                        left_columns.get_column_from_alias(col, only_one=True)
                        for col in self._using
                    ]
                    new_metadata = left_columns + right_columns

                    # we estimate the cardinality of the left table to inform the
                    # batch size for the joins. Cardinality here is the ratio of
                    # unique values in the set. Although we're working it out, we'll
                    # refer to this as an estimate because it may be different per
                    # page of data - we're assuming it's not very different.
                    cols = pyarrow_ops.columns_to_array_denulled(
                        page, left_join_columns
                    )
                    if page.num_rows > 0:
                        card = len(numpy.unique(cols)) / page.num_rows
                    else:
                        card = 0
                    batch_size = calculate_batch_size(card)

                # we break this into small chunks otherwise we very quickly run into memory issues
                for batch in page.to_batches(max_chunksize=batch_size):

                    # blocks don't have column_names, so we need to wrap in a table
                    batch = pyarrow.Table.from_batches([batch], schema=page.schema)

                    new_page = pyarrow_ops.inner_join(
                        self._right_table, batch, right_join_columns, left_join_columns
                    )
                    new_page = new_metadata.apply(new_page)
                    yield new_page

        elif self._on:

            right_columns = Columns(self._right_table)

            right_null_columns, self._right_table = Columns.remove_null_columns(
                self._right_table
            )

            left_columns = None

            for page in left_node.execute():

                if left_columns is None:
                    left_columns = Columns(page)
                    left_join_columns, right_join_columns = get_columns(
                        self._on, left_columns, right_columns
                    )

                    left_null_columns, page = Columns.remove_null_columns(page)
                    new_metadata = right_columns + left_columns

                    # ensure the types are compatible for joining by coercing numerics
                    self._right_table = arrow.coerce_columns(
                        self._right_table, right_join_columns
                    )

                page = arrow.coerce_columns(page, left_join_columns)

                # do the join
                # This uses the cjoin / HASH JOIN / legacy join
                new_page = page.join(
                    self._right_table,
                    keys=left_join_columns,
                    right_keys=right_join_columns,
                    join_type="inner",
                    coalesce_keys=False,
                )

                # update the metadata
                new_page = Columns.restore_null_columns(
                    left_null_columns + right_null_columns, new_page
                )

                new_page = new_metadata.apply(new_page)
                yield new_page



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Blob Reader Node

This is a SQL Query Execution Plan Node.

This Node reads and parses the data from one of the sample datasets.
"""
import datetime

from typing import Iterable

import pyarrow

from opteryx import samples
from opteryx.exceptions import DatasetNotFoundError
from opteryx.models import Columns, QueryProperties
from opteryx.operators import BasePlanNode


def _normalize_to_types(table):
    """
    Normalize types e.g. all numbers are float64 and dates
    """
    schema = table.schema

    for index, column_name in enumerate(schema.names):
        type_name = str(schema.types[index])
        if type_name in ("date32[day]", "date64", "timestamp[s]", "timestamp[ms]"):
            schema = schema.set(
                index, pyarrow.field(column_name, pyarrow.timestamp("us"))
            )

    return table.cast(target_schema=schema)


def _get_sample_dataset(dataset, alias, end_date):
    # we do this like this so the datasets are not loaded into memory unless
    # they are going to be used
    sample_datasets = {
        "$satellites": samples.satellites,
        "$planets": samples.planets,
        "$astronauts": samples.astronauts,
        "$no_table": samples.no_table,
    }
    dataset = dataset.lower()
    if dataset in sample_datasets:
        table = sample_datasets[dataset](end_date)
        table = _normalize_to_types(table)
        table = Columns.create_table_metadata(
            table=table,
            expected_rows=table.num_rows,
            name=dataset,
            table_aliases=[alias],
        )
        return table
    raise DatasetNotFoundError(f"Dataset not found `{dataset}`.")


class InternalDatasetNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        """
        The Blob Reader Node is responsible for reading the relevant blobs
        and returning a Table/Relation.
        """
        super().__init__(properties=properties)

        self._alias = config["alias"]
        self._dataset = config["dataset"]

        today = datetime.datetime.utcnow().date()
        self._start_date = config.get("start_date", today)
        self._end_date = config.get("end_date", today)

    @property
    def config(self):  # pragma: no cover
        if self._alias:
            return f"{self._dataset} => {self._alias}"
        return f"{self._dataset}"

    @property
    def name(self):  # pragma: no cover
        return "Sample Dataset Reader"

    def execute(self) -> Iterable:

        pyarrow_page = _get_sample_dataset(self._dataset, self._alias, self._end_date)
        self.statistics.rows_read += pyarrow_page.num_rows
        self.statistics.bytes_processed_data += pyarrow_page.nbytes
        self.statistics.columns_read += len(pyarrow_page.column_names)

        schema = pyarrow_page.schema

        for index, column_name in enumerate(schema.names):
            type_name = str(schema.types[index])
            if type_name in ("date32[day]", "date64", "timestamp[s]", "timestamp[ms]"):
                schema = schema.set(
                    index,
                    pyarrow.field(
                        name=column_name,
                        type=pyarrow.timestamp("us"),
                        metadata=pyarrow_page.field(column_name).metadata,
                    ),
                )
        pyarrow_page = pyarrow_page.cast(target_schema=schema)

        yield pyarrow_page



try:
    from .sqloxide import parse_sql
except ImportError as e:
    print(e)
    if str(e) != "PyO3 modules may only be initialized once per interpreter process":
        raise e

__all__ = ["parse_sql"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is the Query Planner, it is responsible for creating the physical execution plan
from the SQL statement provided by the user. It does this is a multistage process
as per the below.

 ┌────────────┐  SQL    ┌─────────────────┐
 │            ├────────►│ Parser + Lexer  │
 │            |         └─────────────────┘
 │  Query     │  AST    ┌─────────────────┐
 │   Planner  ├────────►| Binder          │
 │            |         └─────────────────┘
 │            │  AST    ┌─────────────────┐
 │            ├────────►│ Logical Planner │
 │            |         └─────────────────┘
 │            │  Plan   ┌─────────────────┐
 │            ├────────►│ Optimizer       │
 └────────────┘         └─────────────────┘
       │
       ▼
    Executor
"""
from opteryx.config import config
from opteryx.exceptions import SqlError, ProgrammingError
from opteryx.managers.planner import binder
from opteryx.managers.planner.logical import logical_planner
from opteryx.managers.planner.optimizer import run_optimizer
from opteryx.managers.planner.temporal import extract_temporal_filters
from opteryx.models import QueryProperties
from opteryx.third_party import sqloxide


class QueryPlanner:
    def __init__(
        self, *, statement: str = "", cache=None, ast=None, properties=None, qid=None
    ):

        # if it's a byte string, convert to an ascii string
        if isinstance(statement, bytes):
            statement = statement.decode()
        self.raw_statement = statement

        if properties is None:
            self.properties = QueryProperties(qid, config)
            self.properties.cache = cache

            # we need to deal with the temporal filters before we use sqloxide
            if statement is not None:
                (
                    self.statement,
                    self.properties.temporal_filters,
                ) = extract_temporal_filters(statement)
            else:
                self.statement = statement
        else:
            self.properties = properties

        self.ast = ast
        self.logical_plan = None
        self.physical_plan = None

    def parse_and_lex(self):
        """
        Parse & Lex the SQL into an Abstract Syntax Tree (AST)

        We use sqlparser-rs to create an AST from the SQL. More information is
        available in the documentation for that library, but very broadly it
        will tokenize (split into words) the SQL, parse, and then assign meaning
        to each of the tokens, lex, to build the AST.

        The AST is just another representation of the SQL provided by the user,
        when we have the AST we know we probably have valid SQL, but not if we
        can actually execute it.
        """
        try:
            yield from sqloxide.parse_sql(self.statement, dialect="mysql")
            # MySQL Dialect allows identifiers to be delimited with ` (backticks) and
            # identifiers to start with _ (underscore) and $ (dollar sign)
            # https://github.com/sqlparser-rs/sqlparser-rs/blob/main/src/dialect/mysql.rs
        except ValueError as exception:  # pragma: no cover
            raise SqlError(exception) from exception

    def bind_ast(self, ast, parameters):
        return binder.bind_ast(ast, parameters, self.properties)

    def create_logical_plan(self, ast):
        return logical_planner.create_plan(ast, self.properties)

    def optimize_plan(self, plan):
        if self.properties.enable_optimizer:
            return run_optimizer(plan, self.properties)
        return plan

    def execute(self, plan):
        return plan.execute()

    def test_paramcount(self, asts, params):
        """count the number of Placeholders and compare to the number of params"""

        def _inner(node):
            # walk the tree counting Placeholders
            if isinstance(node, list):
                return sum(_inner(n) for n in node)
            if isinstance(node, dict):
                if "Value" in node:
                    if "Placeholder" in node["Value"]:
                        return 1
                return sum(_inner(v) for v in node.values())
            return 0

        found_params = _inner(asts)
        if found_params > len(params):
            raise ProgrammingError(
                "Incorrect number of bindings supplied. Fewer parameters were found than placeholders."
            )
        if found_params < len(params):
            raise ProgrammingError(
                "Incorrect number of bindings supplied. Fewer placeholders are provided than parameters."
            )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This implements an in-memory cache.

We use an LRU-K(2) to maintain the size of the cache, this is a variation of the naive
LRU algorithm.

We're using a dictionary and moving items to the top of the dictionary
when it's accessed. This relies on Python dictionaries being ordered.
"""

import io

from opteryx.managers.kvstores import BaseKeyValueStore
from opteryx.utils.lru_2 import LRU2


class InMemoryKVStore(BaseKeyValueStore):

    slots = ("_lru2",)

    def __init__(self, **kwargs):
        """
        Parameters:
            size: int (optional)
                The maximim number of items maintained in the cache.
        """
        size = int(kwargs.get("size", 50))
        self._lru2 = LRU2(size=size)

    def get(self, key):
        value = self._lru2.get(key)
        if value:
            return io.BytesIO(value)
        return None

    def set(self, key, value):
        value.seek(0, 0)
        ret = self._lru2.set(key, value.read())
        value.seek(0, 0)
        return ret

    def contains(self, keys):
        return list(set(keys).intersection(self._lru2.keys))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Distinct Node

This is a SQL Query Execution Plan Node.

This Node eliminates duplicate records.
"""
from typing import Iterable

from pyarrow import Table, concat_tables

from opteryx.exceptions import SqlError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.third_party.pyarrow_ops import drop_duplicates


class DistinctNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._distinct = config.get("distinct", True)

    @property
    def config(self):  # pragma: no cover
        return ""

    @property
    def greedy(self):  # pragma: no cover
        return True

    @property
    def name(self):  # pragma: no cover
        return "Distinction"

    def execute(self) -> Iterable:

        if len(self._producers) != 1:
            raise SqlError(f"{self.name} on expects a single producer")

        data_pages = self._producers[0]  # type:ignore
        if isinstance(data_pages, Table):
            data_pages = (data_pages,)

        if self._distinct:
            yield drop_duplicates(concat_tables(data_pages.execute(), promote=True))
            return
        yield from data_pages



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

# import pyximport

from pathlib import Path

# pyximport.install()  # build any pyx files

from opteryx import config
from opteryx.connection import Connection
from opteryx.connectors import register_store
from opteryx.version import __version__


__author__: str = "@joocer"

apilevel = "1.0"  # pylint: disable=C0103
threadsafety = 0  # pylint: disable=C0103
paramstyle = "qmark"  # pylint: disable=C0103


def connect(*args, **kwargs):
    """define the opteryx.connect function"""
    return Connection(*args, **kwargs)


# python-dotenv allows us to create an environment file to store secrets. If
# there is no .env it will fail gracefully.
try:
    import dotenv  # type:ignore
except ImportError:  # pragma: no cover
    dotenv = None  # type:ignore

env_path = Path(".") / ".env"

#  deepcode ignore PythonSameEvalBinaryExpressiontrue: false +ve, values can be different
if env_path.exists() and (dotenv is None):  # pragma: no cover  # nosemgrep
    # using a logger here will tie us in knots
    print("`.env` file exists but `dotEnv` not installed.")
elif dotenv is not None:  # pragma: no cover
    dotenv.load_dotenv(dotenv_path=env_path)


# Try to increase the priority of the application
if not config.DISABLE_HIGH_PRIORITY and hasattr(os, "nice"):  # pragma: no cover
    nice_value = os.nice(0)
    try:
        os.nice(-20 + nice_value)
        print(f"Process priority set to {os.nice(0)}.")
    except PermissionError:
        print(f"Cannot update process priority. Currently set to {nice_value}.")

# Log resource usage
if not config.DISABLE_RESOURCE_LOGGING:  # pragma: no cover
    from opteryx.utils.resource_monitor import ResourceMonitor



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Limit Node

This is a SQL Query Execution Plan Node.

This Node returns up to a specified number of tuples.
"""
from typing import Iterable

from pyarrow import Table

from opteryx.exceptions import SqlError
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.utils import arrow


class LimitNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self.limit = config.get("limit")

    @property
    def name(self):  # pragma: no cover
        return "Limitation"

    @property
    def config(self):  # pragma: no cover
        return str(self.limit)

    def execute(self) -> Iterable:

        if len(self._producers) != 1:
            raise SqlError(f"{self.name} on expects a single producer")

        data_pages = self._producers[0]  # type:ignore
        if isinstance(data_pages, Table):
            data_pages = (data_pages,)

        yield arrow.limit_records(data_pages.execute(), limit=self.limit)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## bind variables
## bind temporal ranges
## (bind statistics)

import datetime
import decimal

from typing import Iterable

import numpy

from opteryx.exceptions import ProgrammingError, SqlError
from opteryx.models import QueryProperties


def _build_literal_node(value):
    if value is None:
        return {"Value": "Null"}
    if isinstance(value, (bool)):
        # boolean must be before numeric
        return {"Value": {"Boolean": value}}
    if isinstance(value, (str)):
        return {"Value": {"SingleQuotedString": value}}
    if isinstance(value, (int, float, decimal.Decimal)):
        return {"Value": {"Number": [value, False]}}
    if isinstance(value, (numpy.datetime64)):
        return {"Value": {"SingleQuotedString": value.item().isoformat()}}
    if isinstance(value, (datetime.date, datetime.datetime)):
        return {"Value": {"SingleQuotedString": value.isoformat()}}


def variable_binder(node, parameter_set, properties, query_type):
    """Walk the AST replacing 'Placeholder' nodes, this is recursive"""
    # Replace placeholders with parameters.
    # We do this after the AST has been parsed to remove any chance of the
    # parameter affecting the meaning of any of the other tokens - i.e. to
    # eliminate this feature being used for SQL injection.
    if isinstance(node, list):
        return [variable_binder(i, parameter_set, properties, query_type) for i in node]
    if isinstance(node, dict):
        if "Value" in node:
            if "Placeholder" in node["Value"]:
                # fmt:off
                if len(parameter_set) == 0:
                    raise ProgrammingError("Incorrect number of bindings supplied."
                    " More placeholders are provided than parameters.")
                placeholder_value = parameter_set.pop(0)
                return _build_literal_node(placeholder_value)
                # fmt:on
        # replace @variables
        if query_type != "SetVariable" and "Identifier" in node:
            token_name = node["Identifier"]["value"]
            if token_name[0] == "@":
                if token_name not in properties.variables:  # pragma: no cover
                    raise SqlError(f"Undefined variable found in query `{token_name}`.")
                variable_value = properties.variables.get(token_name)
                return _build_literal_node(variable_value.value)
        return {
            k: variable_binder(v, parameter_set, properties, query_type)
            for k, v in node.items()
        }
    # we're a leaf
    return node


def temporal_range_binder(ast, filters, cache):
    if isinstance(ast, (list)):
        return [temporal_range_binder(node, filters, cache) for node in ast]
    if isinstance(ast, (dict)):
        node_name = next(iter(ast))
        if node_name == "Table":
            temporal_range = filters.pop(0)
            ast["Table"]["start_date"] = temporal_range[1]
            ast["Table"]["end_date"] = temporal_range[2]
            ast["Table"]["cache"] = cache
            return ast
        if "table_name" in ast:
            temporal_range = filters.pop(0)
            ast["table_name"][0]["start_date"] = temporal_range[1]
            ast["table_name"][0]["end_date"] = temporal_range[2]
            ast["table_name"][0]["cache"] = cache
            return ast
        if "ShowCreate" in ast:
            temporal_range = filters.pop(0)
            ast["ShowCreate"]["start_date"] = temporal_range[1]
            ast["ShowCreate"]["end_date"] = temporal_range[2]
            ast["ShowCreate"]["cache"] = cache
            return ast
        return {k: temporal_range_binder(v, filters, cache) for k, v in ast.items()}
    return ast


def statistics_binder(ast):
    return ast


def bind_ast(ast, parameters: Iterable = None, properties: QueryProperties = None):
    """
    Bind physical information to the AST

    This includes the following activities
    - Replacing placeholders with the parameters
    - Adding temporal range information to relations
    """
    query_type = next(iter(ast))

    bound_ast = ast.copy()
    bound_ast = variable_binder(bound_ast, parameters, properties, query_type)
    if query_type in ("Query", "Explain", "ShowColumns", "ShowCreate"):
        bound_ast = temporal_range_binder(
            bound_ast, list(properties.temporal_filters), properties.cache
        )
        bound_ast = statistics_binder(bound_ast)

    return bound_ast



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Eliminate Negations

Type: Heuristic
Goal: Reduce complexity
"""
from opteryx import operators
from opteryx.managers.expression import ExpressionTreeNode, NodeType


def eliminate_negations(plan, properties):
    """
    This action aims to remove steps from the execution of an expression by eliminating
    negations. This is triggered by a NOT in an expression, when this is observed we
    look at what it is negating to see if we can rewrite it so the NOT isn't required.

    This takes the form of removing chained NOTs - this is not actually expected in
    actual use - and seeing if we can invert a condition -
        e.g. 'NOT a = b' == 'a != b'
    """

    # Operations safe to invert.
    inversions: dict = {
        "Eq": "NotEq",
        "NotEq": "Eq",
        "Gt": "LtEq",
        "GtEq": "Lt",
        "Lt": "GtEq",
        "LtEq": "Gt",
    }

    def update_expression_tree(node):
        """
        Walk a expression tree collecting all the nodes of a specified type.
        """
        # this is the main work of this action
        if node.token_type == NodeType.NESTED:
            return update_expression_tree(node.centre)
        if node.token_type == NodeType.NOT:
            centre_node = node.centre
            # a straight forward eliminate NOT(NOT(x))
            if centre_node.token_type == NodeType.NOT:
                return update_expression_tree(centre_node.centre)
            # do we have an invertable operator
            if (
                centre_node.token_type == NodeType.COMPARISON_OPERATOR
                and centre_node.value in inversions
            ):
                centre_node.value = inversions[centre_node.value]
                return update_expression_tree(centre_node)
        # below here is generic to walk the tree
        node.left = None if node.left is None else update_expression_tree(node.left)
        node.centre = (
            None if node.centre is None else update_expression_tree(node.centre)
        )
        node.right = None if node.right is None else update_expression_tree(node.right)
        if node.parameters:
            node.parameters = [
                parameter
                if not isinstance(parameter, ExpressionTreeNode)
                else update_expression_tree(parameter)
                for parameter in node.parameters
            ]
        return node

    # find the in-scope nodes (WHERE AND HAVING)
    selection_nodes = plan.get_nodes_of_type(operators.SelectionNode)

    # killer questions - if any aren't met, bail
    if selection_nodes is None:
        return plan

    # HAVING and WHERE are selection nodes
    for nid in selection_nodes:
        # get the node from the node_id
        operator = plan.get_operator(nid)
        operator.filter = update_expression_tree(operator.filter)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Page Defragment Node

This is a SQL Query Execution Plan Node.

    Orignally implemented to test if datasets have any records as they pass through
    the DAG, this function normalizes the number of bytes per page.

    This is to balance two competing demands:
        - operate in a low memory environment, if the pages are too large they may
          cause the process to fail.
        - operate quickly, if we spend our time doing SIMD on pages with few records
          we're not working as fast as we can.

    The low-water mark is 75% of the target size, less than this we look to merge
    pages together. This is more common following the implementation of projection
    push down, one column doesn't take up a lot of memory so we consolidate tens of
    pages into a single page.

    The high-water mark is 199% of the target size, more than this we split the page.
    Splitting at a size any less than this will end up with pages less that the target
    page size.

"""
from typing import Iterable
import pyarrow
from opteryx.exceptions import SqlError

from opteryx.operators import BasePlanNode

PAGE_SIZE = 64 * 1024 * 1024  # 64Mb
HIGH_WATER: float = 1.99  # Split pages over 199% of PAGE_SIZE
LOW_WATER: float = 0.75  # Merge pages under 75% of PAGE_SIZE


class PageDefragmentNode(BasePlanNode):
    @property
    def name(self):  # pragma: no cover
        return "Page Defragment"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:

        data_pages = self._producers[0]  # type:ignore
        if isinstance(data_pages, pyarrow.Table):
            data_pages = (data_pages,)

        # we can disable this function in the properties
        if not self.properties.enable_page_defragmentation:
            yield from data_pages
            return

        row_counter = 0
        collected_rows = None
        at_least_one_page = False

        for page in data_pages.execute():

            if page.num_rows > 0:

                # add what we've collected before to the table
                if collected_rows:  # pragma: no cover
                    self.statistics.page_merges += 1
                    page = pyarrow.concat_tables([collected_rows, page], promote=True)
                    collected_rows = None

                # work out some stats about what we have
                page_bytes = page.nbytes
                page_records = page.num_rows

                # if we're more than double the target size, let's do something
                if page_bytes > (PAGE_SIZE * HIGH_WATER):  # pragma: no cover
                    average_record_size = page_bytes / page_records
                    new_row_count = int(PAGE_SIZE / average_record_size)
                    row_counter += new_row_count
                    self.statistics.page_splits += 1
                    yield page.slice(offset=0, length=new_row_count)
                    at_least_one_page = True
                    collected_rows = page.slice(offset=new_row_count)
                # if we're less than 75% of the page size, save hold what we have so
                # far and go collect the next page
                elif page_bytes < (PAGE_SIZE * LOW_WATER):
                    collected_rows = page
                # otherwise the page size is okay so we can emit the current page
                else:
                    row_counter += page_records
                    yield page
                    at_least_one_page = True
            elif not at_least_one_page:
                # we have to emit something to the next step, but don't emit
                # multiple empty pages
                yield page
                at_least_one_page = True

        # if we're at the end and haven't emitted all the records, emit them now
        if collected_rows:
            row_counter += collected_rows.num_rows
            yield collected_rows



# datetime_truncate
# ===================
# This module truncates a datetime object to the level of precision that you specify,
# making everything higher than that zero (or one for day and month).
#
# It is based on PostgreSQL's DATE_TRUNC.
#
# Originally from https://github.com/mediapop/datetime_truncate
#
# ------------------------------------------------------------------------------------
#
# This has been updated for Opteryx:
# - Unused functions removed and others renamed
# - the parameter order to match the order in POSTGRES
# - mod is used to truncate sub week values (01-01-1970 is a Thursday so would require
#   additional steps to use this approach)


from datetime import datetime, timezone
from opteryx.utils import dates

__all__ = [
    "date_trunc",
]

NUMERIC_PERIODS = {
    "second": 1,
    "minute": 60,
    "hour": 3600,
    "day": 86400,
}

# fmt:off
PERIODS = {
    "month": dict(microsecond=0, second=0, minute=0, hour=0, day=1),
    "year": dict(microsecond=0, second=0, minute=0, hour=0, day=1, month=1),
}
ODD_PERIODS = { "week", "quarter" }
# fmt:on


def _truncate_week(dt):
    """
    Truncates a date to the first day of an ISO 8601 week, i.e. monday.

    :params dt: an initialized datetime object
    :return: `dt` with the original day set to monday
    :rtype: :py:mod:`datetime` datetime object

    This has been rewritten for Opteryx, it has been moved to the `mod`
    approach, but requires an additional step because 01-JAN-1970 is a
    Thursday.
    """
    from datetime import timedelta

    #   I can't work out why this implementation doesn't work
    #    weeks = dt.timestamp() // 604800
    #    return datetime(1970,1,5, tzinfo=timezone.utc) + timedelta(weeks=weeks)

    dt = date_trunc("day", dt)
    return dt - timedelta(days=dt.isoweekday() - 1)


def _truncate_quarter(dt):
    """
    Truncates the datetime to the first day of the quarter for this date.

    :params dt: an initialized datetime object
    :return: `dt` with the month set to the first month of this quarter
    :rtype: :py:mod:`datetime` datetime object
    """
    dt = date_trunc("month", dt)

    month: int = dt.month
    if 1 <= month <= 3:
        return dt.replace(month=1)
    if 4 <= month <= 6:
        return dt.replace(month=4)
    if 7 <= month <= 9:
        return dt.replace(month=7)
    if 10 <= month <= 12:
        return dt.replace(month=10)
    return None


def date_trunc(truncate_to: str, dt):
    """
    Truncates a dt to have the values with higher precision than
    the one set as `truncate_to` as zero (or one for day and month).

    Possible values for `truncate_to`:

    * second
    * minute
    * hour
    * day
    * week (iso week i.e. to monday)
    * month
    * quarter
    * year

    Examples::

       >>> truncate(datetime(2012, 12, 12, 12), 'day')
       datetime(2012, 12, 12)
       >>> truncate(datetime(2012, 12, 14, 12, 15), 'quarter')
       datetime(2012, 10, 1)
       >>> truncate(datetime(2012, 3, 1), 'week')
       datetime(2012, 2, 27)

    :params dt: an initialized datetime object
    :params truncate_to: The highest precision to keep its original data.
    :return: datetime with `truncated_to` as the highest level of precision
    :rtype: :py:mod:`dt` datetime object
    """
    # convert acceptable non datetime values to datetime
    dt = dates.parse_iso(dt)

    if not isinstance(truncate_to, str):
        truncate_to = truncate_to[0]  # [#325]

    # Added for Opteryx - this improves performance approximately
    # 33% for these items
    if truncate_to in NUMERIC_PERIODS:
        seconds = dt.timestamp()
        seconds -= seconds % NUMERIC_PERIODS[truncate_to]
        return datetime.fromtimestamp(seconds, tz=timezone.utc)

    if truncate_to in PERIODS:
        return dt.replace(**PERIODS[truncate_to])

    if truncate_to == "week":
        return _truncate_week(dt)
    if truncate_to == "quarter":
        return _truncate_quarter(dt)
    raise ValueError(
        f"DATE_TRUNC not valid. Valid periods: {', '.join(list(PERIODS.keys()) + list(ODD_PERIODS))}"
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
import datetime

from typing import Any


@dataclass
class QueryProperties:
    """
    Hints and properties to use when executing queries.
    """

    read_only_properties = ("variables", "cache", "temporal_filters", "date", "qid")

    def __init__(self, qid, config: dict = None):

        # this is empty unless it's set as part of the query
        self.variables: dict[str, Any] = {}

        # fmt:off
        if config is None:
            config = {}
        # query parameters - these can be overridden on a per-query basis

        # use the query optimizer
        self.enable_optimizer:bool = True
        # The maximum input frame size for JOINs
        self.internal_batch_size: int = int(config.get("INTERNAL_BATCH_SIZE", 500))
        # The maximum number of records to create in a CROSS JOIN frame
        self.max_join_size: int = int(config.get("MAX_JOIN_SIZE", 10000))
        # Approximate Page Size
        self.page_size: int = config.get("PAGE_SIZE", 64 * 1024 * 1024)
        # Internally split and merge pages
        self.enable_page_defragmentation: bool = True

        # cost values go here:
        #    costs are the approximate number of seconds to perform an action
        #    1 million times

        # fmt:on

        self.temporal_filters: list = []
        self.date = datetime.datetime.utcnow().date()
        self.cache = None
        self.qid = qid



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Distinct Node

This is a SQL Query Execution Plan Node.

This Node is used to evaluate the filters on SHOW queries e.g.: 

- SHOW COLUMNS
- SHOW VARIABLES

It expects the filter to be applied to the 'name' column.
"""
from typing import Iterable

from opteryx.models import Columns, QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.exceptions import SqlError


class ColumnFilterNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._filter = config.get("filter", True)

    @property
    def config(self):  # pragma: no cover
        return ""

    @property
    def greedy(self):  # pragma: no cover
        return True

    @property
    def name(self):  # pragma: no cover
        return "Select Columns"

    def execute(self) -> Iterable:

        if len(self._producers) != 1:
            raise SqlError(f"{self.name} on expects a single producer")

        data_pages = self._producers[0]  # type:ignore

        selection = None
        columns = None

        for page in data_pages.execute():
            if selection is None:
                columns = Columns(page)
                selection = list(dict.fromkeys(columns.filter(self._filter)))
            page = page.select(selection)
            yield columns.apply(page)
